'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = execute;
var isNode = typeof process !== 'undefined';

var strategyInstance = void 0;

/**
 * Description 根据传入的请求配置发起请求并进行预处理
 * @param url
 * @param option
 * @param {*} acceptType json | text | blob
 * @param strategy
 */
function execute(url) {
  var option = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
  var acceptType = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : 'json';
  var strategy = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : {};

  if (!url) {
    throw new Error('地址未定义');
  }

  var promise = void 0;

  strategyInstance = strategy;

  if (strategy.Jsonp) {
    // Jsonp 只能是 Get 请求，并且不能带函数
    promise = fetch(url);
  } else {
    // 这里判断是否为 Node 环境，是 Node 环境则设置环境变量
    if (typeof process !== 'undefined') {
      // 避免 HTTPS 错误
      process.env.NODE_TLS_REJECT_UNAUTHORIZED = '0';
    }

    //构建fetch请求
    promise = fetch(url, option);
  }

  // 添加基本的处理逻辑
  promise = promise.then(function (response) {
    return _checkStatus(response, acceptType);
  }, function (error) {
    throw error;
  }).then(function (response) {
    // 根据不同的数据类型启用不同的解析方式
    if (acceptType === 'json') {
      return _parseJSON(response);
    } else if (acceptType === 'blob') {
      return _parseBlob(response);
    } else {
      return _parseText(response);
    }
  }, function (error) {
    throw error;
  });

  // 以高阶函数的方式封装 Promise 对象

  return _decorate(promise);
}

/**
 * @function 检测返回值的状态
 * @param response
 * @param acceptType
 * @returns {*}
 */
function _checkStatus(response, acceptType) {
  if (response.status >= 200 && response.status < 300 || response.status === 0) {
    return response;
  } else {
    // 封装错误对象
    var error = new Error(JSON.stringify({
      status: response.status,
      statusText: response.statusText
    }));

    error.response = response;

    throw error;
  }
}

/**
 * @function 解析返回值中的Response为JSON形式
 * @param response
 * @returns {*}
 */
function _parseJSON(response) {
  if (!!response) {
    return response.json();
  } else {
    return {};
  }
}

/**
 * @function 解析TEXT性质的返回
 * @param response
 * @returns {*}
 */
function _parseText(response) {
  if (!!response) {
    return response.text();
  } else {
    return '';
  }
}

/**
 * @function 解析二进制流性质的返回
 * @param response
 * @returns {*}
 */
function _parseBlob(response) {
  if (!!response) {
    if (isNode) {
      return response;
    } else {
      return response.blob();
    }
  } else {
    return null;
  }
}

/**
 * @function 将原始的 Promise 进行封装
 * @private
 * @param initialpromise
 */
function _decorate(initialpromise) {
  var abortFunction = void 0;

  // 默认 60 秒过时
  var timeout = 0;

  var abortablePromise = new Promise(function (resolve, reject) {
    // 闭包方式传递对象
    abortFunction = function abortFunction() {
      reject(new Error('Abort or Timeout'));
    };
  });

  var promise = Promise.race([initialpromise, abortablePromise]);

  // 挂载放弃函数
  promise.abort = abortFunction;

  // 挂载 pipe 函数
  /**
   * @param idOrWriteStream
   * @param fs 如果是 Node 环境，需要额外传入 fs 对象
   * @returns {Promise.<TResult>}
   */
  promise.pipe = function (idOrWriteStream, fs) {
    return promise.then(function (resOrBlobData) {
      // 判断是否为 Node 环境
      if (!isNode) {
        document.querySelector(idOrWriteStream).src = URL.createObjectURL(resOrBlobData);
      } else {
        if (!fs) {
          console.error('fs is null!');
          return;
        }

        // 如果为 Node 环境，则写入到磁盘中
        var dest = fs.createWriteStream(idOrWriteStream);

        // 将数据利用 pipe 写入到文件中
        resOrBlobData.body.pipe(dest);
      }
    }, function (error) {
      throw error;
    });
  };

  // 定义 timeout 对象
  Object.defineProperty(promise, 'timeout', {
    set: function set(ts) {
      if (ts = +ts) {
        timeout = ts;
        setTimeout(abortFunction, ts);
      }
    },
    get: function get() {
      return timeout;
    }
  });

  if (strategyInstance.hasOwnProperty('timeout')) {
    promise.timeout = strategyInstance.timeout;
  }

  return promise;
}